/* hoststat.h - host status database
 * vix 01feb92 [written]
 */

#if !defined(lint) && !defined(LINT)
static char rcsid[] = "$Id: hoststat.c,v 1.5 1993/09/07 07:03:05 vixie Exp vixie $";
#endif /*lint*/

#include "sendmail.h"

#include <assert.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/file.h>

#ifndef S_ISREG
# define	S_ISREG( mode )		(((mode) & S_IFMT) == S_IFREG)
#endif
#ifndef S_ISDIR
# define	S_ISDIR( mode )		(((mode) & S_IFMT) == S_IFDIR)
#endif

char *hsRoot = ".hoststat";

/* int
 * hs_isdir(pathname, createflag)
 *
 * function:
 *	check to see if `pathname' exists.  create it if nec'y and if
 *	`createflag' is set.
 *
 * return:
 *	0: `pathname' existed or was successfully created
 *	-1: otherwise
 */
static int
hs_isdir(pathname, createflag)
	char *pathname;
	int createflag;
{
	struct stat statbuf;
	int ret;

	if (0 > (ret = stat(pathname, &statbuf))) {
		if (errno == ENOENT && createflag)
			ret = mkdir(pathname, 0775);
		return (ret);
	}
	if (!S_ISDIR(statbuf.st_mode)) {
		errno = ENOTDIR;
		return (-1);
	}
	return (0);
}


/* int
 * hs_path(host, path, createflag)
 *
 * function:
 *	given `host', convert from a.b.c to $hsRoot/c./b./a, putting
 *	the result into `path'.  if `createflag' is set, intervening
 *	directories will be created as needed.
 *
 * return:
 *	0: success
 *	-1: failure
 */
static int
hs_path(host, path, createflag)
	const char *host;
	char *path;
	int createflag;
{
	char t_host[MAXHOSTNAMELEN];
	char *elem, *p, *x, ch;
	int ret = 0;

	strcpy(t_host, host);

	/* delete any trailing dots from the hostname.
	 * leave 'elem' pointing at the \0.
	 */
	elem = t_host + strlen(t_host);
	while (elem > t_host && elem[-1] == '.')
		*--elem = '\0';

	strcpy(path, hsRoot);
	p = path + strlen(path);
	/* warning: this loop depends on being able to set a pointer to
	 * an address one character below an array and detect this with
	 * the '>=' operator.  K&R says not to do this, but we're lazy.
	 * it will not work on segmented machines, but since even the
	 * 386 runs in linear-address-space mode these days, we aren't
	 * worried.  not that we worried, anyway.  if it bites you, you have
	 * our apologies and please send us your context diffs :-)...
	 */
	while (elem > t_host
	    && (0 <= (ret = hs_isdir(path, createflag))))
	{
		elem--;
		while (elem >= t_host && *elem != '.')
			elem--;
		*p++ = '/';
		for (x = elem+1; (ch = *x++) && (ch != '.'); NULL) {
			if (isupper(ch))
				ch = tolower(ch);
			if (ch == '|')
				*p++ = '|';	/* | -> || */
			else if (ch == '/')
				ch = '|';	/* / -> | */
			*p++ = ch;
		}
		if (elem >= t_host)
			*p++ = '.';
		*p = '\0';
	}
	return (ret);
}


/* int
 * hs_purge(hostname)
 *
 * function:
 *	given `hostname', compute the path to its status file and delete that
 *	file.  then march upward to hsRoot, deleting all empty directories.
 *
 * return:
 *	0: success
 *	-1: failure (check errno)
 */
int
hs_purge(hostname)
	const char *hostname;
{
	char pathname[MAXPATHLEN];
	char *sentinel, *element;
	struct stat statbuf;

	if (hs_path(hostname, pathname, 0) < 0)
		return (-1);
	if (unlink(pathname) < 0)
		return (-1);

	sentinel = pathname + strlen(hsRoot);
	assert(*sentinel == '/');

	for (element = strrchr(pathname, '/');
	     element && element > sentinel;
	     NULL) {
		*element = '\0';
		if (rmdir(pathname) < 0) {
			if (errno == ENOENT || errno == EEXIST)
				return (0);	/*no more progress to be made*/
#ifdef ENOTEMPTY
			if (errno == ENOTEMPTY)
				return (0);	/* wierd BSDism */
#endif
			return (-1);
		}
		do { element--; } while (element>sentinel && *element != '/');
	}

	return (0);
}


/* FILE *
 * hs_fopen(hostname, createflag)
 *
 * function:
 *	given `hostname', convert from a.b.c to $hsRoot/c./b./a, creating
 *	directories "c." and "b." if nec'y (and if createflag is set), and
 *	open "a" for read/update ("r+") or for create/read/write ("w+")
 *	if `createflag' is set.
 *
 * return:
 *	file handle if file could be opened or created, or NULL if it
 *	did not exist and could not be created.
 *
 * notes:
 *	we use trailing "."'s on the intermediate directories because we
 *	otherwise have a collision between "@apple.com" and
 *	"@appleshare.apple.com".
 */
static FILE *
hs_fopen(hostname, createflag)
	const char *hostname;
	int createflag;
{
	char pathname[MAXPATHLEN];
	FILE *f;
	int ret;

	ret = hs_path(hostname, pathname, createflag);
	if (ret < 0)
		return (NULL);
	if (f = fopen(pathname, "r+"))
		return (f);
	if (createflag)
		return (fopen(pathname, "w+"));
	return (NULL);
}


/* struct hoststat *
 * hs_parsef(f, hostname)
 *
 * function:
 *	given open filehandle `f', read/parse/rewind the file and using
 *	the parsed info plus the passed `hostname' plus the result of an
 *	fstat() (for the mtime), construct a (struct hoststat).
 *
 * return:
 *	a pointer to the constructed (struct hoststat) or NULL if the
 *	parse failed for any reason.
 */
static struct hoststat *
hs_parsef(f, hostname)
	FILE *f;
	char *hostname;
{
	static struct hoststat ret;
	struct stat statbuf;
	char ch, *s, *d;

	for (d = ret.hostname, s = hostname; (ch = *s++) != '\0'; NULL) {
		if (ch == '|') {
			if (s[1] == '|')
				continue;	/* || -> | */
			ch = '/';		/* | -> / */
		}
		*d++ = ch;
	}
	*d = '\0';

	ret.exitstat = -1;
	ret.errno = -1;
	ret.lf = NULL;
	fstat(fileno(f), &statbuf);
	ret.mtime = statbuf.st_mtime;
	rewind(f);
	while (EOF != (ch = getc(f))) {
		switch (ch) {
		case 'S':
			fscanf(f, "%d\n", &ret.exitstat);
			break;
		case 'E':
			fscanf(f, "%d\n", &ret.errno);
			break;
		default:
			/* forward compatibility through ignorance */
			break;
		}
	}
	if (ret.exitstat != -1 && ret.errno != -1)
		return (&ret);
	return (NULL);
}


/* struct hoststat *
 * hs_find(hostname, lockflag)
 *
 * function:
 *	given `hostname' "a.b.c", look for a file $hsRoot/c/b/a, parse it.
 *	if `lockflag' is set, try to flock(LOCK_EX) the file before parsing.
 *	if we lock it, we return with it still locked; an explicit call to
 *	hs_unlock() must be made to unlock the file.
 *	
 * return:
 *	pointer to a static constructed (struct hoststat), or NULL if:
 *		the file didn't exist and couldn't be created, or
 *		the file couldn't be locked and `lockflag' was set, or
 *		the parse failed.
 *	note that the constructed (struct hoststat)'s `lf' member will
 *	be NULL if no lock was desired, or non-NULL if it represents an
 *	open, locked file.
 */
struct hoststat *
hs_find(hostname, lockflag)
	const char *hostname;
	int lockflag;
{
	int save_errno = errno;
	struct hoststat *ret = NULL;
	FILE *f = hs_fopen(hostname, FALSE);

	if (!f)
		goto done;

	if (lockflag && (0 > flock(fileno(f), LOCK_EX|LOCK_NB))) {
		fclose(f);
		goto done;
	}

	if (!(ret = hs_parsef(f, hostname))) {
		fclose(f);
		goto done;
	}

	if (lockflag)
		ret->lf = f;
	else
		fclose(f);
 done:
	errno = save_errno;
	return (ret);
}


/* struct hoststat *
 * hs_enter(hs, hostname, exitstat, h_errno)
 *
 * function:
 *	create or rewrite a file to contain the (hostname,exitstat,h_errno)
 *	tuple passed as arguments.  if `hs' is non-NULL, it represents an
 *	open and locked file in its `lf' member, and its other members are
 *	to be updated from the passed-in tuple.  if `hs' is NULL, there is
 *	no open file and we will make use of a static (struct hoststat).
 *
 * return:
 *	the updated (struct hoststat), either the same one passed in as `hs'
 *	or our own static one.  the returned hs->lf will always be non-NULL
 *	indicating an open and locked hoststat file.  return NULL if the file
 *	could not be created and/or locked.
 */
struct hoststat *
hs_enter(hs, hostname, exitstat, h_errno)
	struct hoststat *hs;
	const char *hostname;
	int exitstat;
	int h_errno;
{
	int save_errno = errno;
	static struct hoststat my_hs;
	FILE *f;

	/* use our static hs if one was not provided */
	if (!hs) {
		hs = &my_hs;
		hs->lf = NULL;
	}

	/* update fields in hs */
	hs->errno = h_errno;
	hs->exitstat = exitstat;
	strcpy(hs->hostname, hostname);

	/* open file if it is not already open */
	if (!hs->lf) {
		hs->lf = hs_fopen(hostname, TRUE);
		if (!hs->lf) {
			hs = NULL;
			goto done;
		}
		if (0 > flock(fileno(hs->lf), LOCK_EX|LOCK_NB)) {
			fclose(hs->lf);
			hs = NULL;
			goto done;
		}
	}

	ftruncate(fileno(hs->lf), 0);
	rewind(hs->lf);
	fprintf(hs->lf, "E%d\n", h_errno);
	fprintf(hs->lf, "S%d\n", exitstat);
	fflush(hs->lf);
 done:
	errno = save_errno;
	return (hs);
}


/* int
 * hs_traverse(action, pathname, lockflag)
 *
 * function:
 *	recursively find all the hoststat files in `pathname' (if non-NULL)
 *	or in hsRoot (defined in this module).  if `lockflag' is set, try
 *	to lock each one and set our local hs->lf according to our success.
 *	call (*action)(&local_hs) for each hoststat file found.  if (*action)
 *	returns a value less than zero, we instantly close all our open files
 *	and dirs and return that less-than-zero value.
 *
 * return:
 *	see above.
 */
int
hs_traverse(action, pathname, lockflag)
	int (*action)();
	const char *pathname;
	int lockflag;
{
	int save_errno = errno;
	struct stat statbuf;
	DIR *d;
	int ret = 0;

	if (!pathname)
		pathname = hsRoot;
	if (0 <= stat(pathname, &statbuf)) {
		if (S_ISDIR(statbuf.st_mode) && (d = opendir(pathname))) {
# ifdef pyr
			struct direct *e;
# else !pyr
			struct dirent *e;
# endif
			char newpath[MAXPATHLEN];
			char *newptr;

			strcpy(newpath, pathname);
			newptr = newpath + strlen(newpath);
			*newptr++ = '/';
			while (e = readdir(d)) {
				if (e->d_name[0] == '.')
					continue;
				strncpy(newptr, e->d_name, e->d_namlen);
				newptr[e->d_namlen] = '\0';
				ret = hs_traverse(action, newpath, lockflag);
				if (ret < 0)
					break;
			}
			closedir(d);
		} else if (S_ISREG(statbuf.st_mode)) {
			FILE *f;

			f = fopen(pathname, "r");
			if (f) {
				struct hoststat *hs;
				char hostname[MAXHOSTNAMELEN], *p, *h, *save;

				/* reconstruct hostname.  note that the
				 * skipping of the leading component is
				 * a side-effect of the way the loops
				 * exit.  hsRoot had better be a single
				 * component or the result of this hack
				 * will be ugly.
				 */
				h = hostname;
				save = NULL;
				while (p = rindex(pathname, '/')) {
					char *x = p+1;

					while (*x && *x != '/')
						*h++ = *x++;
					/* for a long time we did not put dots
					 * at the end of our intermediate path
					 * components (subdirs).  it's cheap to
					 * continue to support old databases by
					 * appending the dot if there isn't one
					 */
					if (h > hostname && h[-1] != '.')
						*h++ = '.';
					if (save)
						*save = '/';
					*p = '\0';
					save = p;	/* there was a '/' */
				}
				if (save)
					*save = '/';
				if (h > hostname && h[-1] == '.')
				       	h--;
				*h = '\0';

				if (hs = hs_parsef(f, hostname)) {
					if (lockflag &&
					    !flock(fileno(f), LOCK_EX|LOCK_NB))
						hs->lf = f;
					ret = (*action)(hs);
				}
				fclose(f);	/* implicit unlock if locked */
			}
		}
	}
	errno = save_errno;
	return (ret);
}


/* void
 * hs_unlock(hs)
 *
 * function:
 *	given `hs', if its `lf' member is non-NULL, assume that it
 *	represents an open and locked hoststat file, and fclose() it
 *	(which also un-flock()'s it).
 *
 * return:
 *	void.
 */
void
hs_unlock(hs)
	struct hoststat *hs;
{
	int save_errno = errno;

	if (hs && hs->lf) {
		fclose(hs->lf);
		hs->lf = NULL;
	}
	errno = save_errno;
}
